import numpy as np
import numpy.typing as npt
from dataclasses import dataclass
from typing import NamedTuple, Optional


def convert(preds):
    if len(preds.shape) == 1:
        return preds
    return np.argmax(preds, axis=1).ravel()


class FairnessMeasure:
    def __init__(
        self,
        y_unique: npt.NDArray[int],
        s_unique: npt.NDArray[int],
        y: npt.NDArray[int],
        s: npt.NDArray[int],
    ):
        self.y_unique = y_unique
        self.s_unique = s_unique
        self.init_C(y, s)
        self.init_P(y, s)


class EqualizedOdds(FairnessMeasure):
    r"""The function implements the equal odds fairness function. A model :math:`h_θ` is fair for Equalized Odds when
    the probability of predicting the correct label is independent of the sensitive attribute.

    Args:
        y_unique (npt.ndarray[int]): all unique labels in all label space.
        s_unique (npt.ndarray[int]): all unique protected attributes in all protected attribute space.
        y (npt.ndarray[int]): all label space
        s (npt.ndarray[int]): all protected attribute space
    """

    def init_C(self, y, s):
        n_groups = self.y_unique.shape[0] * self.s_unique.shape[0]
        self.C = np.zeros((n_groups, n_groups))

        indices = np.ravel_multi_index(
            (
                [l for l in self.y_unique for r in self.s_unique],
                [r for l in self.y_unique for r in self.s_unique],
            ),
            (self.y_unique.shape[0], self.s_unique.shape[0]),
        )

        for i in indices:
            for j in indices:
                l, r = np.unravel_index(
                    i, (self.y_unique.shape[0], self.s_unique.shape[0])
                )
                lp, rp = np.unravel_index(
                    j, (self.y_unique.shape[0], self.s_unique.shape[0])
                )
                if l == lp:
                    self.C[i, j] = np.sum(s[y == lp] == rp) / np.sum(y == lp)
        self.C = self.C - np.eye(n_groups)

    def init_P(self, y, s):
        self.P = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))
        for l in self.y_unique:
            for r in self.s_unique:
                self.P[l, r] = np.mean(np.logical_and(y == l, s == r))

    def groupwise(self, preds, y, s):
        preds = convert(preds)

        groupwise_fairness = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))

        for l in self.y_unique:
            reference_rate = np.mean(preds[y == l] == l)
            for r in self.s_unique:
                mask = np.logical_and(y == l, s == r)
                groupwise_fairness[l, r] = np.mean(preds[mask] == l) - reference_rate

        return groupwise_fairness


class AccuracyParity(FairnessMeasure):
    r"""The function implements the accuracy parity fairness function.
    A model :math:`h_θ` is fair for Accuracy Parity when theprobability of being correct is independent of the sensitive attribute.

    Args:
        y_unique (npt.ndarray[int]): all unique labels in all label space.
        s_unique (npt.ndarray[int]): all unique protected attributes in all protected attribute space.
        y (npt.ndarray[int]): all label space
        s (npt.ndarray[int]): all protected attribute space
    """

    def init_C(self, y, s):
        n_groups = self.y_unique.shape[0] * self.s_unique.shape[0]
        self.C = np.zeros((n_groups, n_groups))
        indices = np.ravel_multi_index(
            (
                [l for l in self.y_unique for r in self.s_unique],
                [r for l in self.y_unique for r in self.s_unique],
            ),
            (self.y_unique.shape[0], self.s_unique.shape[0]),
        )

        for j in indices:
            lp, rp = np.unravel_index(
                j, (self.y_unique.shape[0], self.s_unique.shape[0])
            )
            self.C[:, j] = np.mean(s == rp)

        self.C = (
            self.C
            - np.eye(n_groups)
            - np.eye(n_groups, k=self.s_unique.shape[0])
            - np.eye(n_groups, k=-self.s_unique.shape[0])
        ) / self.y_unique.shape[0]

    def init_P(self, y, s):
        self.P = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))
        for r in self.s_unique:
            self.P[:, r] = np.mean(s == r)

    def groupwise(self, preds, y, s):
        preds = convert(preds)

        groupwise_fairness = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))

        reference_rate = np.mean(preds == y)
        for r in self.s_unique:
            mask = s == r
            groupwise_fairness[:, r] = np.mean(preds[mask] == y[mask]) - reference_rate

        return groupwise_fairness


class EqualityOpportunity:
    r"""The function implements the accuracy parity fairness function.
    A model :math:`h_θ` is fair for Equality of Opportunity when the probability of predicting the
    correct label is independent of the sensitive attribute for a given subset of labels called the desirable outcomes

    Args:
        y_unique (npt.ndarray[int]): all unique labels in all label space.
        s_unique (npt.ndarray[int]): all unique protected attributes in all protected attribute space.
        y (npt.ndarray[int]): all label space
        s (npt.ndarray[int]): all protected attribute space
        y_desirable (npt.ndarray[int]): the label for which the fairness needs to be enforced.
    """

    def __init__(
        self,
        y_unique: npt.NDArray[int],
        s_unique: npt.NDArray[int],
        y: npt.NDArray[int],
        s: npt.NDArray[int],
        y_desirable: npt.NDArray[int],
    ):
        self.y_desirable = y_desirable
        self.y_unique = y_unique
        self.s_unique = s_unique
        self.y = y
        self.s = s
        self.init_C(self.y, self.s)
        self.init_P(self.y, self.s)

    def init_C(self, y, s):
        n_groups = self.y_unique.shape[0] * self.s_unique.shape[0]
        self.C = np.zeros((n_groups, n_groups))

        indices = np.ravel_multi_index(
            (
                [l for l in self.y_unique for r in self.s_unique],
                [r for l in self.y_unique for r in self.s_unique],
            ),
            (self.y_unique.shape[0], self.s_unique.shape[0]),
        )

        for i in indices:
            for j in indices:
                l, r = np.unravel_index(
                    i, (self.y_unique.shape[0], self.s_unique.shape[0])
                )
                lp, rp = np.unravel_index(
                    j, (self.y_unique.shape[0], self.s_unique.shape[0])
                )
                if l == lp:
                    self.C[i, j] = np.sum(s[y == lp] == rp) / np.sum(y == lp)
        self.C = self.C - np.eye(n_groups)
        for i in indices:
            l, r = np.unravel_index(i, (self.y_unique.shape[0], self.s_unique.shape[0]))
            if l not in self.y_desirable:
                self.C[i, :] = 0

    def init_P(self, y, s):
        self.P = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))
        for l in self.y_unique:
            for r in self.s_unique:
                self.P[l, r] = np.mean(np.logical_and(y == l, s == r))

    def groupwise(self, preds, y, s):
        preds = convert(preds)

        groupwise_fairness = np.zeros((self.y_unique.shape[0], self.s_unique.shape[0]))

        for l in self.y_desirable:
            reference_rate = np.mean(preds[y == l] == l)
            for r in self.s_unique:
                mask = np.logical_and(y == l, s == r)

                groupwise_fairness[l, r] = np.mean(preds[mask] == l) - reference_rate

        return groupwise_fairness


@dataclass
class FairnessSetupArguments:
    fairness_function_name: str
    y_unique: np.asarray
    s_unique: np.asarray
    all_train_y: np.asarray
    all_train_s: np.asarray
    y_desirable: Optional[np.asarray] = None


def setup_fairness_function(fairness_setup: FairnessSetupArguments):
    if fairness_setup.fairness_function_name == "equal_odds":
        fairness_function = EqualizedOdds(
            y_unique=fairness_setup.y_unique,
            s_unique=fairness_setup.s_unique,
            y=fairness_setup.all_train_y,
            s=fairness_setup.all_train_s,
        )
    elif fairness_setup.fairness_function_name == "equal_opportunity":
        fairness_function = EqualityOpportunity(
            y_unique=fairness_setup.y_unique,
            s_unique=fairness_setup.s_unique,
            y_desirable=fairness_setup.y_desirable,
            y=fairness_setup.all_train_y,
            s=fairness_setup.all_train_s,
        )
    elif fairness_setup.fairness_function_name == "accuracy_parity":
        fairness_function = AccuracyParity(
            y_unique=fairness_setup.y_unique,
            s_unique=fairness_setup.s_unique,
            y=fairness_setup.all_train_y,
            s=fairness_setup.all_train_s,
        )
    else:
        raise NotImplementedError
    return fairness_function



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fairgrad', 'fairgrad.torch']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx>=5.1.1,<6.0.0', 'numpy>=1.21', 'torch>=1.0']

setup_kwargs = {
    'name': 'fairgrad',
    'version': '0.1.3',
    'description': '',
    'long_description': '# FairGrad: Fairness Aware Gradient Descent\n[![Documentation Status](https://readthedocs.org/projects/pip/badge/)](https://fairgrad.readthedocs.io/en/latest/)\n[![PyPI version](https://badge.fury.io/py/fairgrad.svg)](https://badge.fury.io/py/fairgrad)\n<a href="https://arxiv.org/abs/2206.10923"><img src="http://img.shields.io/badge/Paper-PDF-red.svg"></a>\n\nFairGrad, is an easy to use general purpose approach to enforce fairness for gradient descent based methods. \n\n# Getting started: \nYou can get ```fairgrad``` from pypi, which means it can be easily installed via ```pip```:\n```\npip install fairgrad\n```\n\n# Documentation\nThe documenation can be found at [read the docs](https://fairgrad.readthedocs.io/en/latest/index.html)\n\n# Example usage \nTo use fairgrad simply replace your pytorch cross entropy loss with fairgrad cross entropy loss. \nAlongside, regular pytorch cross entropy arguments, it expects following extra arguments.\n\n```\ny_train (np.asarray[int], Tensor, optional): All train example\'s corresponding label\ns_train (np.asarray[int], Tensor, optional): All train example\'s corresponding sensitive attribute. This means if there\n        are 2 sensitive attributes, with each of them being binary. For instance gender - (male and female) and\n        age (above 45, below 45). Total unique sentive attributes are 4.\nfairness_measure (string): Currently we support "equal_odds", "equal_opportunity", and "accuracy_parity".\nepsilon (float, optional): The slack which is allowed for the final fairness level.\nfairness_rate (float, optional): Parameter which intertwines current fairness weights with sum of previous fairness rates.\n```\n\n```python\n# Note this is short snippet. One still needs to models and iterators.\n# Full worked out example is available here - @TODO\n\nfrom fairgrad.torch import CrossEntropyLoss\n\n# define cross entropy loss \ncriterion = CrossEntropyLoss(fairness_related_meta_data=fairness_related_meta_data)\n\n# Train loop\n\nfor inputs, labels, protected_attributes in train_iterator:\n    model.train()\n    optimizer.zero_grad()\n    output = model(inputs)\n    loss = criterion(output, labels, protected_attributes, mode=\'train\')\n    loss.backward()\n    optimizer.step()\n```\n\n# Citation\n```\n@article{maheshwari2022fairgrad,\n  title={FairGrad: Fairness Aware Gradient Descent},\n  author={Maheshwari, Gaurav and Perrot, Micha{\\"e}l},\n  journal={arXiv preprint arXiv:2206.10923},\n  year={2022}\n}\n```\n',
    'author': 'gmaheshwari',
    'author_email': 'gaurav.maheshwari@inria.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import warnings

from torch.nn.modules.distance import PairwiseDistance
from torch.nn.modules.module import Module
from torch.nn import functional as F
from torch.nn import _reduction as _Reduction

import torch
import numpy as np
from torch import Tensor
import numpy.typing as npt
from fairgrad.fairness_functions import *
from typing import Callable, Optional, Union


def get_fairness_function(fairness_function, all_train_y, all_train_s):
    fairness_setup_arguments = FairnessSetupArguments(
        fairness_function_name=fairness_function,
        y_unique=np.unique(all_train_y),
        s_unique=np.unique(all_train_s),
        all_train_y=all_train_y,
        all_train_s=all_train_s,
        y_desirable=np.asarray([1]),
    )

    return setup_fairness_function(fairness_setup_arguments)


class _Loss(Module):
    reduction: str

    def __init__(self, size_average=None, reduce=None, reduction: str = "mean") -> None:
        super(_Loss, self).__init__()
        if size_average is not None or reduce is not None:
            self.reduction = _Reduction.legacy_get_string(size_average, reduce)
        else:
            self.reduction = reduction


class _WeightedLoss(_Loss):
    def __init__(
        self,
        weight: Optional[Tensor] = None,
        size_average=None,
        reduce=None,
        reduction: str = "mean",
    ) -> None:
        super(_WeightedLoss, self).__init__(size_average, reduce, reduction)
        self.register_buffer("weight", weight)


class CrossEntropyLoss(_WeightedLoss):
    r"""This is an extension of the CrossEntropyLoss provided by pytorch. Please check pytorch documentation
    for understanding the cross entropy loss. Here, we augment the cross entropy to enforce fairness. The
    exact algorithm can be found in Fairgrad paper <https://arxiv.org/abs/2206.10923>.

    Args:
        weight (Tensor, optional): a manual rescaling weight given to each class.
            If given, has to be a Tensor of size `C`
        size_average (bool, optional): Deprecated (see :attr:`reduction`). By default,
            the losses are averaged over each loss element in the batch. Note that for
            some losses, there are multiple elements per sample. If the field :attr:`size_average`
            is set to ``False``, the losses are instead summed for each minibatch. Ignored
            when reduce is ``False``. Default: ``True``
        ignore_index (int, optional): Specifies a target value that is ignored
            and does not contribute to the input gradient. When :attr:`size_average` is
            ``True``, the loss is averaged over non-ignored targets.
        reduce (bool, optional): Deprecated (see :attr:`reduction`). By default, the
            losses are averaged or summed over observations for each minibatch depending
            on :attr:`size_average`. When :attr:`reduce` is ``False``, returns a loss per
            batch element instead and ignores :attr:`size_average`. Default: ``True``
        reduction (string, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will
            be applied, ``'mean'``: the weighted mean of the output is taken,
            ``'sum'``: the output will be summed. Note: :attr:`size_average`
            and :attr:`reduce` are in the process of being deprecated, and in
            the meantime, specifying either of those two args will override
            :attr:`reduction`. Default: ``'mean'``
        y_train (np.asarray[int], Tensor, optional): All train example's corresponding label
        s_train (np.asarray[int], Tensor, optional): All train example's corresponding sensitive attribute. This means if there
            are 2 sensitive attributes, with each of them being binary. For instance gender - (male and female) and
            age (above 45, below 45). Total unique sentive attributes are 4.
        fairness_measure (string): Currently we support "equal_odds", "equal_opportunity", and "accuracy_parity".
        epsilon (float, optional): The slack which is allowed for the final fairness level.
        fairness_rate (float, optional): Parameter which intertwines current fairness weights with sum of previous fairness rates.



    Examples::

        >>> input = torch.randn(10, 5, requires_grad=True)
        >>> target = torch.empty(10, dtype=torch.long).random_(2)
        >>> s = torch.empty(10, dtype=torch.long).random_(2) # protected attribute
        >>> loss = CrossEntropyLoss(y_train = target, s_train = s, fairness_measure = 'equal_odds')
        >>> output = loss(input, target, s, mode='train')
        >>> output.backward()
    """
    __constants__ = ["ignore_index", "reduction"]
    ignore_index: int

    def __init__(
        self,
        weight: Optional[Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
        y_train: Optional[Union[npt.NDArray[int], Tensor]] = None,
        s_train: Optional[Union[npt.NDArray[int], Tensor]] = None,
        fairness_measure: Optional[str] = None,
        epsilon: Optional[float] = 0.0,
        fairness_rate: Optional[float] = 0.01,
    ) -> None:
        super(CrossEntropyLoss, self).__init__(weight, size_average, reduce, reduction)
        self.ignore_index = ignore_index
        if y_train is not None and s_train is not None and fairness_measure is not None:
            self.fairness_flag = True
        else:
            self.fairness_flag = False
            warnings.warn(
                "Fairgrad mechanism is not employed as either y_train or s_train "
                "or fairness_measure is missing. Reverting to regular Cross Entropy Loss"
            )
        if self.fairness_flag:
            # Below keys is necessary
            assert all(y_train >= 0), "labels space must start with 0, and not -1"
            assert all(
                s_train >= 0
            ), "protected label space must start with 0, and not -1"
            self.y_train = y_train
            self.s_train = s_train
            self.epsilon = epsilon
            self.fairness_measure = fairness_measure
            self.fairness_rate = fairness_rate

            if type(self.y_train) == torch.Tensor:
                self.y_train = self.y_train.detach().numpy()
            if type(self.s_train) == torch.Tensor:
                self.s_train = self.s_train.detach().numpy()

            self.y_unique = np.unique(self.y_train)
            self.s_unique = np.unique(self.s_train)
            self.fairness_function = get_fairness_function(
                self.fairness_measure, self.y_train, self.s_train
            )

            # some lists to store all information over time. This in the future would be given as an option so
            # as to save space. @TODO: describe them individually.
            self.step_weights = []
            self.step_loss = []
            self.step_weighted_loss = []
            self.step_accuracy = []
            self.step_fairness = []
            self.cumulative_fairness = torch.zeros(
                (self.y_unique.shape[0], 2 * self.s_unique.shape[0])
            )

    def forward(
        self, input: Tensor, target: Tensor, s: Tensor = None, mode: str = "train"
    ) -> Tensor:
        loss = F.cross_entropy(
            input,
            target,
            weight=self.weight,
            ignore_index=self.ignore_index,
            reduction="none",
        )
        if self.fairness_flag and mode == "train":

            groupwise_fairness = self.fairness_function.groupwise(
                input.detach().numpy(), target.detach().numpy(), s.detach().numpy()
            )

            if self.step_fairness != []:
                groupwise_fairness = np.where(
                    np.isnan(groupwise_fairness),
                    self.step_fairness[-1],
                    groupwise_fairness,
                )
            else:
                groupwise_fairness = np.where(
                    np.isnan(groupwise_fairness), 0, groupwise_fairness
                )

            self.step_fairness.append(groupwise_fairness)

            # Compute the weights
            self.cumulative_fairness = np.clip(
                self.cumulative_fairness
                + self.fairness_rate
                * np.hstack(
                    (
                        groupwise_fairness - self.epsilon,
                        -(groupwise_fairness + self.epsilon),
                    )
                ),
                0,
                None,
            )

            lag_parameters = (
                self.cumulative_fairness[:, : self.s_unique.shape[0]]
                - self.cumulative_fairness[:, self.s_unique.shape[0] :]
            )
            partial_weights = (
                self.fairness_function.C.T.dot(lag_parameters.reshape(-1, 1))
            ).reshape(self.y_unique.shape[0], self.s_unique.shape[0])
            weights = 1 + partial_weights / self.fairness_function.P
            self.step_loss.append(loss)
            self.step_weights.append(weights)
            weighted_loss = loss * torch.tensor((weights)[target, s])
            self.step_weighted_loss.append(weighted_loss.mean())
            if self.reduction == "mean":
                return torch.mean(weighted_loss)
            elif self.reduction == "sum":
                return torch.sum(weighted_loss)
            else:
                return weighted_loss
        if self.reduction == "mean":
            return torch.mean(loss)
        elif self.reduction == "sum":
            return torch.sum(loss)
        else:
            return loss
